/*
 *
 *
 *   bintail.cc
 *
 *   'tail' for binary files.  This little utility will read a file from
 *   the filesystem, and send its contents to standard output, byte-by-
 *   byte.  This is similar to tail(1), but bintail doesn't care about
 *   line endings or 7-bit ASCII, etc.
 *
 *   There are currently some primitive seek commands built in, where you
 *   can move forward or backward in the current stream, by pressing keys
 *   during runtime:
 *
 *   f - move forward 1kB
 *   F - move forward 1MB
 *   W - move forward 10MB
 *   b - move backward 1kB
 *   B - move backward 1MB
 *   0 - (zero) move to start of file
 *   $ - move to end of file
 *
 *   The seek functionality needs to be refined, but it does work.  This
 *   utility is still a work in progress.
 *
 *   The original purpose for this utility was to pipe data from a WAV
 *   file to a process while the file was being recorded by another
 *   unrelated process, in real time.
 *
 *   Copyright (C) 2006 by Matt Roberts, KK5JY.
 *   License: GNU GPL2 (www.gnu.org)
 *   Version: 20060122(c)
 *
 *
 */

#include <string>
#include <iostream>
#include <fstream>

#include <cstdlib>

#include <sys/select.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <unistd.h>
#include <fcntl.h>
#include <cstring>

using namespace std;

// prime pump with 64k
const unsigned int prime = 64 * 1024;

// buffer size
const unsigned int BUFSZ = 1 * 1024;


/*
 *
 *   usage() - display help on how to run the program
 *
 */

void usage (void) {
	cerr << "Usage: bintail <filename>" << endl;
	exit(1);
}


/*
 *
 *  ioerror() - invoked upon fatal I/O error while running I/O functions
 *
 */

void ioerror (const std::string &s, const std::string &m) {
	if (s.empty())
		cerr << "bintail: I/O error: " << m << endl;
	else
		cerr << "bintail: I/O error while tailing '" << s << "': " << m << endl;
	exit(1);
}


/*
 *
 *   badfile() - invoked upon fatal error opening file
 *
 */

void badfile (const std::string &s, const std::string &m) {
	cerr << "bintail: could not open file '" << s << "': " << m << endl;
	exit(1);
}


/*
 *
 *   fsize(fd) - determine size of file associated with fd
 *
 */

off64_t fsize (int fd) {
	// fstat64 will store its info here
	struct stat64 sb;

	// measure current file size
	if (fstat64(fd, &sb) != 0)
		ioerror("", strerror(errno));

	// return st_size field
	return sb.st_size;
}


/*
 *
 *   ltell(fd) - return current I/O position within file 'fd'
 *
 */

off64_t ltell (int fd) {
	// result value
	off64_t pos = 0;

	// measure current file size
	pos = lseek64(fd, 0, SEEK_CUR);
	if (errno != 0)
		ioerror("", strerror(errno));

	// return st_size field
	return pos;
}


/*
 *
 *   absseek() - seek to requested byte offset within the file
 *
 */

off64_t absseek (int fd, off64_t howfar) {
	// determine where the edge of the file is
	off64_t edge  = fsize(fd);

	// position to which to seek
	off64_t where = howfar;

	// check for overflow (and clip if request would overrun edge)
	if (where < 0)
		where = 0;
	if (where > edge)
		where = edge;

	// DEBUG:
	// cerr << "bintail: absseek(" << where << "): seeking to offset " << where << endl;

	// seek
	return lseek64(fd, where, SEEK_SET);
}


/*
 *
 *   relseek() - seek to position relative to current position; 'clip' the
 *               request to file limits to prevent overrun
 *
 */

off64_t relseek (int fd, off64_t howfar) {
	// determine current file location
	off64_t now   = ltell(fd);

	// determine where the edge of the file is
	off64_t edge  = fsize(fd);

	// compute new location
	off64_t where = now + howfar;
	
	// check for overflow (and clip if request would overrun edge)
	if (where < 0)
		where = 0;
	if (where > edge)
		where = edge;

	// DEBUG:
	// cerr << "bintail: relseek(" << now << " -> " << howfar << "): seeking to offset " << where << endl;

	// seek
	return lseek64(fd, where, SEEK_SET);
}


/*
 *
 *   reopen(...) - reopen a file descriptor, preserving location
 *
 */

int reopen (int fd, const std::string &fn, mode_t md) {
	off64_t pos = ltell(fd);
	close(fd);
	fd = open(fn.c_str(), md);
	if (fd < 0)
		badfile(fn, strerror(errno));
	lseek64(fd, pos, SEEK_SET);
	return fd;
} 


/*
 *
 *   main() - entry point
 *
 */

int main (int argc, char **argv) {
	// make sure enough args were given
	if ((argc < 2) || (argc > 3))
		usage();
	
	// open the input file
	int input = open(argv[1], O_RDONLY);
	if (input < 0)
		badfile(argv[1], strerror(errno));

	// flush C++ streams
	cerr.flush();
	cout.flush();

	// setup vars
	fd_set rdset;
	int n = 1, written = 0;
	unsigned char buffer[BUFSZ];
	unsigned char cmdbuf[1024];
	off64_t lastsize = 0; // last reported size of the file
	struct stat64 sb;

	// read starting offset, if given
	off64_t pos = 0; // starting offset (careful!)
	if (argc == 3)
		pos = atol(argv[2]);
	else
		pos = 0;

	// seek to starting offset
	pos = lseek64(input, pos, SEEK_SET);

	// initialize 'sb'
	sb.st_size = 0;

	// do initial 'stat' to determine file size
	if (fstat64(input, &sb) != 0) {
		cerr << "bintail: fstat() failed (" << strerror(errno) << "), tailing may not work." << endl;
		usleep(500000);
	}

	// forever
	while (true) {
		while (sb.st_size == lastsize) {
			// call fstat to measure the file's size
			if (fstat64(input, &sb) != 0)
				cerr << "bintail: fstat() failed (" << strerror(errno) << "), tailing may not work." << endl;

			// sleep for half a second
			usleep(500000);
		}

		// save the new file size
		lastsize = sb.st_size;

		// save current position, reopen the file
		input = reopen(input, argv[1], O_RDONLY);

		// while data is available, copy from file to stdout
		while (true) {
			// setup file descriptor sets
			FD_ZERO(&rdset);
			FD_SET(input, &rdset); // the file
			FD_SET(0,     &rdset); // stdin
			n = input + 1;

			// call select(), to wait for data
			n = select(n, &rdset, 0, 0, 0);

			// do we have an input event?
			if (n > 0) {
				// if input data from file...
				if (FD_ISSET(input, &rdset)) {
					n = read(input, buffer, sizeof(buffer));

					// if bytes were read
					if (n > 0) {
						// loop until all bytes written
						while (n > 0) {
							// attempt to write ALL read bytes to stdout
							written = write(1, buffer, n);

							if (written > 0) {
								n -= written;
							} else if (written < 0) {
								ioerror(argv[1], strerror(errno));
							}
						}
					} else if (n < 0) {
						ioerror(argv[1], strerror(errno));
					} else if (n == 0) {
						break;
					}
				}

				// deal with console commands...
				if (FD_ISSET(0, &rdset)) {
					// use to indicate that we have performed an immediate seek
					bool tainted = false;

					// read commands from stdin
					n = read(0, cmdbuf, sizeof(cmdbuf));
					if (n > 0) {
						// for each command...
						for (int i = 0; i < n; ++i) {
							// act according to the command
							switch (cmdbuf[i]) {
								case 'Q':
								case 'q': exit(1);
								case '0': lseek64(input, 0, SEEK_SET);      tainted = true; break;
								case '$': lseek64(input, -64, SEEK_END);    tainted = true; break;
								case 'f': relseek(input, 1024);             tainted = true; break;
								case 'F': relseek(input, 1024 * 1024);      tainted = true; break;
								case 'W': relseek(input, 10 * 1024 * 1024); tainted = true; break;
								case 'b': relseek(input, -1024);            tainted = true; break;
								case 'B': relseek(input, -(1024 * 1024));   tainted = true; break;
								default: cerr << "bintail: Bad command: " << cmdbuf[i] << endl;
							}
						}
					}

					// if we performed a seek, reopen the file to make sure that
					//    subsequent select() operations perform as we expect
					if (tainted)
						input = reopen(input, argv[1], O_RDONLY);
				}
			}
		}
	}
}

/* EOF: bintail.cc */
